@@ -7,7 +7,7 @@ use File::Which qw( which );
 use File::Spec;
 use Env qw( @PATH );
 
-our $VERSION = '0.02';
+our $VERSION = '0.03';
 
 =head1 NAME
 
@@ -18,7 +18,10 @@ the Rust programming language
 
 Rust:
 
- // compile with: rustc --crate-type dylib add.rs
+ #![crate_type = "dylib"]
+ 
+ // compile with: rustc add.rs
+ 
  #[no_mangle]
  pub extern "C" fn add(a:i32, b:i32) -> i32 {
    a+b
